import random
import operator
from typing import TypedDict, Annotated
from dotenv import load_dotenv

from langchain_core.language_models.chat_models import BaseChatModel
from langchain_core.prompts import PromptTemplate
from langchain_core.output_parsers import StrOutputParser
from langgraph.checkpoint.memory import MemorySaver
from langgraph.types import Send
from langgraph.graph import END, START, StateGraph


load_dotenv()


from ._prompts import _self_discover_prompts as prompts
from ._helpers.logger import logger
from ._helpers.enums import Phase
from ._helpers.validators import validate_params
from ._helpers import stream_response, invoke_graph
from ._config import LLM, Langfuse


class SelfDiscoverState(TypedDict):
    task_description: list[str]
    answer_formats: str
    task_examples: str
    selected_modules: str
    adapted_modules: str
    reasoning_structure: str
    reasoning: Annotated[list[str], operator.add]


class PhaseIIState(TypedDict):
    task_description: str
    answer_formats: str
    reasoning_structure: str


select_prompt = PromptTemplate.from_template(prompts.SELECT_PROMPT)
adapt_prompt = PromptTemplate.from_template(prompts.ADAPT_PROMPT)
implement_prompt = PromptTemplate.from_template(prompts.IMPLEMENT_PROMPT)
reasoning_prompt = PromptTemplate.from_template(prompts.REASONING_PROMPT)


## Phase 1: Discovering Task-Specific Reasoning Structures ##


def select(state: SelfDiscoverState):
    logger.info("Executing SELECT step")
    chain = select_prompt | LLM.model | StrOutputParser()

    result = stream_response(chain, state)

    return {"selected_modules": result}


def adapt(state: SelfDiscoverState):
    logger.info("Executing ADAPT step")
    chain = adapt_prompt | LLM.model | StrOutputParser()

    result = stream_response(chain, state)

    return {"adapted_modules": result}


def implement(state: SelfDiscoverState):
    logger.info("Executing IMPLEMENT step")
    chain = implement_prompt | LLM.model | StrOutputParser()

    result = stream_response(chain, state)

    return {"reasoning_structure": result}


## Phase 2: Reasoning with Task-Specific Reasoning Structures ##


def dummy_node(state: SelfDiscoverState):
    pass


def continue_to_reason(state: SelfDiscoverState):
    logger.info("Sending tasks to REASON step from continue_to_reason")
    return [
        Send(
            "reason",
            {
                "task_description": task_description,
                "answer_formats": state["answer_formats"],
                "reasoning_structure": state["reasoning_structure"],
            },
        )
        for task_description in state["task_description"]
    ]


def reason(state: PhaseIIState):
    logger.debug("Executing REASON step")
    chain = reasoning_prompt | LLM.model | StrOutputParser()

    result = stream_response(chain, state)

    return {"reasoning": [result]}


## Compile graph with memory saver ##


def compile_graph(graph: StateGraph):
    logger.debug("Compiling Self-Discover graph")
    memory = MemorySaver()

    compiled_graph = graph.compile(checkpointer=memory)
    return compiled_graph


## Self-Discover Graph ##


def build_self_discover_graph():
    logger.info("Building self-discover graph")
    graph_builder = StateGraph(SelfDiscoverState)

    ## Phase I
    graph_builder.add_node(select)
    graph_builder.add_node(adapt)
    graph_builder.add_node(implement)

    graph_builder.add_edge(START, "select")
    graph_builder.add_edge("select", "adapt")
    graph_builder.add_edge("adapt", "implement")

    ## Phase II
    graph_builder.add_node(reason)

    graph_builder.add_conditional_edges("implement", continue_to_reason, ["reason"])
    graph_builder.add_edge("reason", END)

    return compile_graph(graph_builder)


## Self-Discover Phase I Graph ##


def build_self_discover_phaseI_graph():
    logger.info("Building self-discover Phase I graph")
    graph_builder = StateGraph(SelfDiscoverState)

    ## Phase I
    graph_builder.add_node(select)
    graph_builder.add_node(adapt)
    graph_builder.add_node(implement)

    graph_builder.add_edge(START, "select")
    graph_builder.add_edge("select", "adapt")
    graph_builder.add_edge("adapt", "implement")

    return compile_graph(graph_builder)


## Self-Discover Phase II Graph ##


def build_self_discover_phaseII_graph():
    logger.info("Building self-discover Phase II graph")
    graph_builder = StateGraph(SelfDiscoverState)

    ## Phase II
    graph_builder.add_node(dummy_node)
    graph_builder.add_node(reason)

    graph_builder.add_edge(START, "dummy_node")
    graph_builder.add_conditional_edges("dummy_node", continue_to_reason, ["reason"])
    graph_builder.add_edge("reason", END)

    return compile_graph(graph_builder)


def call_graph(state, phase, stream):
    logger.info(
        "Calling graph with phase: {}, stream: {}",
        phase,
        stream,
    )

    if phase == Phase.BOTH.value:
        graph = build_self_discover_graph()
    elif phase == Phase.I.value:
        graph = build_self_discover_phaseI_graph()
    elif phase == Phase.II.value:
        graph = build_self_discover_phaseII_graph()
    else:
        logger.error("Invalid phase: {}", phase)
        raise ValueError(
            f"Invalid phase. Must be {Phase.BOTH.value}, {Phase.I.value}, or {Phase.II.value}."
        )

    return invoke_graph(graph, state, Langfuse.CONFIG, stream)


## Function to execute Self-Discover on a series of tasks ##


def self_discover(
    task_description: list[str],
    model: BaseChatModel,
    answer_formats: str,
    reasoning_structure: str = "",
    phase: int = Phase.BOTH.value,
    stream: bool = False,
):
    logger.info("Starting Self-Discover")
    validate_params(task_description, model, answer_formats)

    LLM.model = model

    if phase in (Phase.BOTH.value, Phase.I.value):
        state = {
            "task_description": task_description,
            "task_examples": "\n\n".join(
                random.sample(task_description, min(10, len(task_description)))
            ),
            "answer_formats": answer_formats,
        }
    elif phase == Phase.II.value:
        state = {
            "task_description": task_description,
            "answer_formats": answer_formats,
            "reasoning_structure": reasoning_structure,
        }

    result = call_graph(state, phase, stream)
    logger.debug("Self-Discover result: {}", result.keys())
    return result



# Directly adapted and modified from https://langchain-ai.github.io/langgraph/tutorials/self-discover/self-discover

import re
from typing import Optional, TypedDict

from langchain_core.language_models.chat_models import BaseChatModel
from langchain_core.prompts import PromptTemplate
from langchain_core.output_parsers import StrOutputParser
from langgraph.graph import END, START, StateGraph
from dotenv import load_dotenv

load_dotenv()

from ._prompts import _iself_discover_prompts as prompts
from ._helpers.logger import logger
from ._helpers.validators import validate_params
from ._helpers import stream_response, invoke_graph
from ._config import LLM, Langfuse


## Initial Prompts ##

select_prompt = PromptTemplate.from_template(prompts.SELECT_PROMPT)
adapt_prompt = PromptTemplate.from_template(prompts.ADAPT_PROMPT)

## Reason Prompts ##

# Structured #
structured_planning_prompt = PromptTemplate.from_template(
    prompts.STRUCTURED_PLANNING_PROMPT
)
structured_application_prompt = PromptTemplate.from_template(
    prompts.STRUCTURED_APPLICATION_PROMPT
)

# Unstructured #
unstructured_planning_prompt = PromptTemplate.from_template(
    prompts.UNSTRUCTURED_PLANNING_PROMPT
)
unstructured_application_prompt = PromptTemplate.from_template(
    prompts.UNSTRUCTURED_APPLICATION_PROMPT
)


class ISelfDiscoverState(TypedDict):
    task_description: str
    answer_formats: str
    few_shot_examples: str
    task_description_backup: Optional[str]
    selected_modules: Optional[str]
    adapted_modules: Optional[str]
    reasoning_plan: Optional[str]
    reasoning: Optional[str]


## Corrector nodes for few shot examples ##


def append_few_shot_examples(state: ISelfDiscoverState):
    logger.info("Appending few shot examples to task description")

    task_description = state["task_description"]

    return {
        "task_description": state["task_description"]
        + f"\n\n{prompts.FEW_SHOT_EXAMPLES_PROMPT.format(few_shot_examples_str=state['few_shot_examples'])}",
        "task_description_backup": task_description,
    }


def restore_task_description(state: ISelfDiscoverState):
    logger.info("Restoring original task description before APPLICATION step")

    return {"task_description": state["task_description_backup"]}


## Initial Nodes ##


def select(state: ISelfDiscoverState):
    logger.info("Executing SELECT step")
    chain = select_prompt | LLM.model | StrOutputParser()

    result = stream_response(chain, state)

    return {"selected_modules": result}


def adapt(state: ISelfDiscoverState):
    logger.info("Executing ADAPT step")
    chain = adapt_prompt | LLM.model | StrOutputParser()

    result = stream_response(chain, state)

    return {"adapted_modules": result}  # -> reasoning


## Reason Nodes ##

# Structured #


def structured_planning(state: ISelfDiscoverState):
    logger.info("Executing STRUCTURED PLANNING step")
    chain = structured_planning_prompt | LLM.model | StrOutputParser()

    result = stream_response(chain, state)

    return {"reasoning_plan": result}


def structured_application(state: ISelfDiscoverState):
    logger.info("Executing STRUCTURED APPLICATION step")
    chain = structured_application_prompt | LLM.model | StrOutputParser()

    result = stream_response(chain, state)

    return {"reasoning": result}


# Unstructured #


def unstructured_planning(state: ISelfDiscoverState):
    logger.info("Executing UNSTRUCTURED PLANNING step")
    chain = unstructured_planning_prompt | LLM.model | StrOutputParser()

    result = stream_response(chain, state)

    return {"reasoning_plan": result}


def unstructured_application(state: ISelfDiscoverState):
    logger.info("Executing UNSTRUCTURED APPLICATION step")
    chain = unstructured_application_prompt | LLM.model | StrOutputParser()

    result = stream_response(chain, state)

    return {"reasoning": result}


## Routing Functions ##


def determine_initial_node(state: ISelfDiscoverState):
    if state["few_shot_examples"]:
        return "append_few_shot_examples"
    else:
        return "select"


def determine_pre_application_node(state: ISelfDiscoverState):
    if state["few_shot_examples"]:
        return "restore_task_description"
    else:
        return "application"


def build_iself_discover_graph(structured: bool = False):
    logger.info("Building iSelf-Discover graph")
    graph_builder = StateGraph(ISelfDiscoverState)

    graph_builder.add_node(append_few_shot_examples)
    graph_builder.add_node(restore_task_description)
    graph_builder.add_node(select)
    graph_builder.add_node(adapt)

    if structured:
        graph_builder.add_node("planning", structured_planning)
        graph_builder.add_node("application", structured_application)
    else:
        graph_builder.add_node("planning", unstructured_planning)
        graph_builder.add_node("application", unstructured_application)

    graph_builder.add_conditional_edges(START, determine_initial_node)
    graph_builder.add_edge("append_few_shot_examples", "select")
    graph_builder.add_edge("select", "adapt")
    graph_builder.add_edge("adapt", "planning")
    graph_builder.add_conditional_edges("planning", determine_pre_application_node)
    graph_builder.add_edge("restore_task_description", "application")
    graph_builder.add_edge("application", END)

    logger.debug("Compiling iSelf-Discover graph")
    return graph_builder.compile()


def iself_discover(
    task_description: str,
    model: BaseChatModel,
    answer_formats: str,
    structured: bool = False,
    few_shot_examples_str: str = "",
    stream: bool = False,
):
    logger.info("Starting iSelf-Discover")
    validate_params(task_description, model, answer_formats)

    LLM.model = model

    state = {
        "task_description": task_description,
        "answer_formats": answer_formats,
        "few_shot_examples": few_shot_examples_str,
    }

    graph = build_iself_discover_graph(structured)

    result = invoke_graph(graph, state, Langfuse.CONFIG, stream)
    logger.debug("iSelf-Discover result: {}", result.keys())
    return result



import os

from loguru import logger
from logtail import LogtailHandler
from pyprojroot import here


# handler = LogtailHandler(
#     source_token=os.getenv("BETTERSTACK_TELEMETRY_SOURCE_TOKEN"),
#     host=os.getenv("BETTERSTACK_TELEMETRY_INGESTING_HOST"),
# )


logger.add(here("evals/logs/evaluation.log"))
# logger.add(handler)



from langchain_core.language_models.chat_models import BaseChatModel

from .logger import logger

def validate_params(
    task_description: list[str] | str, model: BaseChatModel, answer_formats: str
):
    logger.debug("Validating parameters")

    if not isinstance(model, BaseChatModel):
        logger.error("Invalid model type: {}", type(model))
        raise ValueError(
            "The model must be an instance of BaseChatModel. Ensure you have defined your LangChain Chat Model properly."
        )

    if isinstance(task_description, str):
        task_description = [task_description]

    if not task_description or not all(
        isinstance(task, str) for task in task_description
    ):
        logger.error("Invalid task_description: {}", task_description)
        raise ValueError("task_description must be a non-empty list of strings.")

    if not isinstance(answer_formats, str):
        logger.error("Invalid answer_formats type: {}", type(answer_formats))
        raise ValueError("answer_formats must be a string.")


from .logger import logger


def stream_response(chain, state):
    result = []

    for chunk in chain.stream(state):
        result.append(chunk)
        # print(chunk, end="", flush=True)
    return "".join(result)


def invoke_graph(graph, state, config, stream):
    logger.info(
        "Invoking graph with state: {}, config: {}, stream: {}",
        state.keys(),
        config,
        stream,
    )
    if stream:
        for s in graph.stream(state, config=config):
            logger.debug(s)

        return graph.get_state(config).values
    else:
        result = graph.invoke(state, config=config)
        return result



from dataclasses import dataclass

from langchain_core.language_models.chat_models import BaseChatModel

# from ._helpers.handlers import langfuse_handler


@dataclass
class LLM:
    model: BaseChatModel


@dataclass
class Langfuse:
    CONFIG = {
        "configurable": {"thread_id": 1}, 
        # "callbacks": [langfuse_handler]
    }


import os

from langfuse.callback import CallbackHandler


# langfuse_handler = CallbackHandler(
#     public_key=os.environ["LANGFUSE_PUBLIC_KEY"],
#     secret_key=os.environ["LANGFUSE_SECRET_KEY"],
#     host=os.environ["LANGFUSE_HOST"],
# )